@@ -8,29 +8,40 @@
  *                                           /\____/
  *                                           \_/__/
  *
- *      Linux console internal mouse driver for event interface
- *      (EVDEV).
+ *      Linux console mouse driver for event interface (EVDEV).
  *
  *      By Annie Testes.
  *
  *      Handles all inputs that fill evdev with informations
  *      about movement and click.
  *
+ *      Hacked for new mouse API by Peter Wang.
+ *
  *      See readme.txt for copyright information.
  */
 
 
-/* linux/input.h also defines KEY_A et al. */
-#define ALLEGRO_NO_KEY_DEFINES
-#include "allegro.h"
+#define ALLEGRO_NO_COMPATIBILITY
+
+#include "allegro5/allegro.h"
 
 #ifdef ALLEGRO_HAVE_LINUX_INPUT_H
 
-#include "allegro/internal/aintern.h"
-#include "allegro/platform/aintunix.h"
-#include "linalleg.h"
+ALLEGRO_DEBUG_CHANNEL("lmseev");
+
+#include "allegro5/internal/aintern.h"
+#include "allegro5/internal/aintern_mouse.h"
+#include "allegro5/platform/aintunix.h"
+#include "allegro5/platform/aintlnx.h"
+
+#ifdef ALLEGRO_RASPBERRYPI
+#include "allegro5/internal/aintern_system.h"
+#include "allegro5/internal/aintern_display.h"
+#include "allegro5/internal/aintern_vector.h"
+#endif
 
 #include <stdio.h>
+#include <string.h>
 #include <sys/ioctl.h>
 #include <linux/input.h>
 
@@ -39,6 +50,31 @@
 #define PREFIX_E                "al-evdev ERROR: "
 
 
+typedef struct AL_MOUSE_EVDEV
+{
+   ALLEGRO_MOUSE parent;
+   int fd;
+   ALLEGRO_MOUSE_STATE state;
+} AL_MOUSE_EVDEV;
+
+
+/* the one and only mouse object */
+static AL_MOUSE_EVDEV the_mouse;
+
+
+
+/* forward declarations */
+static void process_new_data(void *unused);
+static void process_event(struct input_event *event);
+static void handle_button_event(unsigned int button, bool is_down);
+static void handle_axis_event(int dx, int dy, int dz);
+static void generate_mouse_event(unsigned int type,
+                                 int x, int y, int z,
+                                 int dx, int dy, int dz,
+                                 unsigned int button);
+bool _al_evdev_set_mouse_range(int x1, int y1, int x2, int y2);
+
+
 
 /*
  * Tools data
@@ -132,7 +168,7 @@ typedef struct AXIS {
 /* in_to_screen:
  *  maps an input absolute position to a screen position
  */
-static int in_to_screen(AL_CONST AXIS *axis, int v)
+static int in_to_screen(const AXIS *axis, int v)
 {
    return (((v-axis->in_min) * OUT_RANGE(*axis)) / IN_RANGE(*axis)) + axis->out_min;
 }
@@ -195,7 +231,7 @@ static void get_axis_value(int fd, AXIS *axis, int type)
 static int has_event(int fd, unsigned short type, unsigned short code)
 {
    const unsigned int len = sizeof(unsigned long)*8;
-   const unsigned int max = MAX(EV_MAX, MAX(KEY_MAX, MAX(REL_MAX, MAX(ABS_MAX, MAX(LED_MAX, MAX(SND_MAX, FF_MAX))))));
+   const unsigned int max = _ALLEGRO_MAX(EV_MAX, _ALLEGRO_MAX(KEY_MAX, _ALLEGRO_MAX(REL_MAX, _ALLEGRO_MAX(ABS_MAX, _ALLEGRO_MAX(LED_MAX, _ALLEGRO_MAX(SND_MAX, FF_MAX))))));
    unsigned long bits[(max+len-1)/len];
    if (ioctl(fd, EVIOCGBIT(type, max), bits)) {
      return (bits[code/len] >> (code%len)) & 1;
@@ -227,15 +263,6 @@ static AXIS y_axis;
 static AXIS z_axis;
 
 
-/*
- * Previous button status
- */
-static int button_left = 0;
-static int button_right = 0;
-static int button_middle = 0;
-
-
-
 
 /*
  * Initialization functions
@@ -244,28 +271,28 @@ static int button_middle = 0;
 /* init_axis:
  *  initialize an AXIS structure, from the device and the config file
  */
-static void init_axis(int fd, AXIS *axis, AL_CONST char *name, AL_CONST char *section, int type)
+static void init_axis(int fd, AXIS *axis, const char *name, const char *section, int type)
 {
+#if 0
    char tmp1[256]; /* config string */
    char tmp2[256]; /* format string */
    char tmp3[256]; /* Converted 'name' */
+#endif
    int abs[5]; /* values given by the input */
    int config_speed;
 
-   uszprintf(tmp1, sizeof(tmp1), uconvert_ascii("ev_min_%s", tmp2), uconvert_ascii(name, tmp3));
-   axis->in_min = get_config_int(section, tmp1, 0);
-   uszprintf(tmp1, sizeof(tmp1), uconvert_ascii("ev_max_%s", tmp2), uconvert_ascii(name, tmp3));
-   axis->in_max = get_config_int(section, tmp1, 0);
+   (void)name;
+   (void)section;
 
-   uszprintf(tmp1, sizeof(tmp1), uconvert_ascii("ev_abs_to_rel_%s", tmp2), uconvert_ascii(name, tmp3));
-   config_speed = get_config_int(section, tmp1, 1);
-   if (config_speed<=0) config_speed = 1;
+   config_speed = 1;
    axis->scale = 1;
 
    /* Ask the input */
    if (ioctl(fd, EVIOCGABS(type), abs)>=0) {
-      if (axis->in_min==0) axis->in_min=abs[1];
-      if (axis->in_max==0) axis->in_max=abs[2];
+      if (axis->in_min==0)
+         axis->in_min=abs[1];
+      if (axis->in_max==0)
+         axis->in_max=abs[2];
       axis->in_abs = abs[0];
       axis->scale = 320.0*config_speed/IN_RANGE(*axis);
    }
@@ -274,8 +301,8 @@ static void init_axis(int fd, AXIS *axis, AL_CONST char *name, AL_CONST char *se
       axis->scale = 1;
    }
 
-   axis->out_min = 0;
-   axis->out_max = 0;
+   axis->out_min = INT_MIN;
+   axis->out_max = INT_MAX;
 
    axis->speed = 1;
    axis->mickeys = 0;
@@ -288,12 +315,8 @@ static void init_axis(int fd, AXIS *axis, AL_CONST char *name, AL_CONST char *se
  */
 static void init_tablet(int fd)
 {
-   char tmp[256];
-   char *mouse_str = uconvert_ascii("mouse", tmp);
-   char tmp2[256];
    int default_abs = default_tool->mode==MODE_ABSOLUTE;
 
-   default_abs = get_config_int(mouse_str, uconvert_ascii("ev_absolute", tmp2), default_abs);
    if (default_abs) {
       default_tool->mode = MODE_ABSOLUTE;
    }
@@ -301,36 +324,23 @@ static void init_tablet(int fd)
       default_tool->mode = MODE_RELATIVE;
    }
 
-   init_axis(fd, &x_axis, "x", mouse_str, ABS_X);
-   init_axis(fd, &y_axis, "y", mouse_str, ABS_Y);
-   init_axis(fd, &z_axis, "z", mouse_str, ABS_Z);
+   init_axis(fd, &x_axis, "x", "mouse", ABS_X);
+   init_axis(fd, &y_axis, "y", "mouse", ABS_Y);
+   init_axis(fd, &z_axis, "z", "mouse", ABS_Z);
 }
 
 
 
 
-/*
- * Driver structure
- */
-static int processor (unsigned char *buf, int buf_size);
-static INTERNAL_MOUSE_DRIVER intdrv = {
-   -1, /* device, fill it later */
-   processor,
-   3  /* num_buttons */
-};
-
-
-
-
 /*
  * Process input functions
  */
 
-/* process_key:
+/* process_key: [fdwatch thread]
  *  process events of type key (button clicks and vicinity events are currently
  *  supported)
  */
-static void process_key(AL_CONST struct input_event *event)
+static void process_key(const struct input_event *event)
 {
    switch (event->code) {
       /* Buttons click
@@ -338,17 +348,17 @@ static void process_key(AL_CONST struct input_event *event)
        * if event->value is 0 the button has just been released */
       case BTN_LEFT: /* Mouse */
       case BTN_TOUCH: /* Stylus */
-         button_left = !!event->value;
+         handle_button_event(1, event->value);
          break;
 
       case BTN_RIGHT: /* Mouse */
       case BTN_STYLUS: /* Stylus */
-         button_right = !!event->value;
+         handle_button_event(2, event->value);
          break;
 
       case BTN_MIDDLE: /* Mouse */
       case BTN_STYLUS2: /* Stylus */
-         button_middle = !!event->value;
+         handle_button_event(3, event->value);
          break;
 
       /* Vicinity events
@@ -364,11 +374,11 @@ static void process_key(AL_CONST struct input_event *event)
       case BTN_TOOL_LENS:
          if (event->value) {
             current_tool = find_tool(event->code);
-            get_axis_value(intdrv.device, &x_axis, ABS_X);
-            get_axis_value(intdrv.device, &y_axis, ABS_Y);
-            get_axis_value(intdrv.device, &z_axis, ABS_Z);
+            get_axis_value(the_mouse.fd, &x_axis, ABS_X);
+            get_axis_value(the_mouse.fd, &y_axis, ABS_Y);
+            get_axis_value(the_mouse.fd, &z_axis, ABS_Z);
 #ifdef ABS_WHEEL  /* absent in 2.2.x */
-            get_axis_value(intdrv.device, &z_axis, ABS_WHEEL);
+            get_axis_value(the_mouse.fd, &z_axis, ABS_WHEEL);
 #endif
          }
          else {
@@ -379,21 +389,48 @@ static void process_key(AL_CONST struct input_event *event)
 }
 
 
+/* [fdwatch thread] */
+static void handle_button_event(unsigned int button, bool is_down)
+{
+   unsigned int event_type;
 
-/* process_rel:
+   if (is_down) {
+      the_mouse.state.buttons |= (1 << (button-1));
+      event_type = ALLEGRO_EVENT_MOUSE_BUTTON_DOWN;
+   }
+   else {
+      the_mouse.state.buttons &=~ (1 << (button-1));
+      event_type = ALLEGRO_EVENT_MOUSE_BUTTON_UP;
+   }
+
+   generate_mouse_event(
+      event_type,
+      the_mouse.state.x, the_mouse.state.y, the_mouse.state.z,
+      0, 0, 0,
+      button);
+}
+                          
+
+
+
+/* process_rel: [fdwatch thread]
  *  process relative events (x, y and z movement are currently supported)
  */
-static void process_rel(AL_CONST struct input_event *event)
+static void process_rel(const struct input_event *event)
 {
    /* The device can send a report when there's no tool */
    if (current_tool!=no_tool) {
       switch (event->code) {
          case REL_X:
             x_axis.out_abs = rel_event(&x_axis, event->value);
+            handle_axis_event(x_axis.mickeys, 0, 0);
+            x_axis.mickeys = 0;
             break;
 
          case REL_Y:
             y_axis.out_abs = rel_event(&y_axis, event->value);
+            handle_axis_event(0, y_axis.mickeys, 0);
+            y_axis.mickeys = 0;
             break;
 
 #ifdef REL_WHEEL  /* absent in 2.2.x */
@@ -401,6 +438,8 @@ static void process_rel(AL_CONST struct input_event *event)
 #endif
          case REL_Z:
             z_axis.out_abs = rel_event(&z_axis, event->value);
+            handle_axis_event(0, 0, z_axis.mickeys);
+            z_axis.mickeys = 0;
             break;
       }
    }
@@ -408,10 +447,11 @@ static void process_rel(AL_CONST struct input_event *event)
 
 
 
-/* process_abs:
+/* process_abs: [fdwatch thread]
  *  process absolute events (x, y and z movement are currently supported)
+ *  TODO: missing handle_axis_event calls
  */
-static void process_abs(AL_CONST struct input_event *event)
+static void process_abs(const struct input_event *event)
 {
    /* The device can send a report when there's no tool */
    if (current_tool!=no_tool) {
@@ -436,99 +476,24 @@ static void process_abs(AL_CONST struct input_event *event)
 
 
 
-/* processor:
- *  Processes the first packet in the buffer, if any, returning the number
- *  of bytes eaten.
- */
-static int processor (unsigned char *buf, int buf_size)
+/* [fdwatch thread] */
+static void handle_axis_event(int dx, int dy, int dz)
 {
-   struct input_event *event;
-
-   if (buf_size < (int)sizeof(struct input_event))
-      return 0; /* not enough data */
-
-   event = (struct input_event*)buf;
-   switch (event->type) {
-      case EV_KEY:
-         process_key(event);
-         break;
-      case EV_REL:
-         process_rel(event);
-         break;
-      case EV_ABS:
-         process_abs(event);
-         break;
-   }
-
-   if (current_tool!=no_tool) {
-      x_axis.out_abs = CLAMP(x_axis.out_min, x_axis.out_abs, x_axis.out_max);
-      y_axis.out_abs = CLAMP(y_axis.out_min, y_axis.out_abs, y_axis.out_max);
+   if (current_tool != no_tool) {
+      x_axis.out_abs = _ALLEGRO_CLAMP(x_axis.out_min, x_axis.out_abs, x_axis.out_max);
+      y_axis.out_abs = _ALLEGRO_CLAMP(y_axis.out_min, y_axis.out_abs, y_axis.out_max);
       /* There's no range for z */
 
-      _mouse_x = x_axis.out_abs;
-      _mouse_y = y_axis.out_abs;
-      _mouse_z = z_axis.out_abs;
-      _mouse_b = button_left + (button_right<<1) + (button_middle<<2);
-      _handle_mouse_input();
+      the_mouse.state.x = x_axis.out_abs;
+      the_mouse.state.y = y_axis.out_abs;
+      the_mouse.state.z = z_axis.out_abs;
+
+      generate_mouse_event(
+         ALLEGRO_EVENT_MOUSE_AXES,
+         the_mouse.state.x, the_mouse.state.y, the_mouse.state.z,
+         dx, dy, dz,
+         0);
    }
-   /* Returns the size of the data used */
-   return sizeof(struct input_event);
-}
-
-
-
-/* analyse_data:
- *  Analyses the given data, returning 0 if it is unparsable, nonzero
- *  if there's a reasonable chance that this driver can work with that
- *  data.
- *  Used in the setup code
- */
-static int analyse_data (AL_CONST char *buffer, int size)
-{
-   struct input_event *event;
-   struct input_event *last=(struct input_event*)(buffer+size);
-
-   if (size < (int)sizeof(struct input_event))
-      return 0; /* not enough data */
-
-   for (event=(struct input_event*)buffer; event<last; event++) {
-     if (event->type >= EV_MAX)
-        return 0;
-
-     switch (event->type) {
-       case EV_KEY:
-         if (event->code>=KEY_MAX)
-           return 0;
-         break;
-
-       case EV_REL:
-         if (event->code>=REL_MAX)
-           return 0;
-         break;
-
-       case EV_ABS:
-         if (event->code>=ABS_MAX)
-           return 0;
-         break;
-
-#ifdef EV_MSC  /* absent in 2.2.x */
-       case EV_MSC:
-         if (event->code>=MSC_MAX)
-           return 0;
-         break;
-#endif
-
-       /* Mouse doesn't handle events of these types */
-       case EV_LED:
-       case EV_SND:
-       case EV_REP:
-#ifdef EV_FF  /* absent in 2.2.x */
-       case EV_FF:
-#endif
-         return 0;
-     }
-   }
-   return event==last;
 }
 
 
@@ -543,13 +508,13 @@ static int open_mouse_device (const char *device_file)
 
    fd = open (device_file, O_RDONLY | O_NONBLOCK);
    if (fd >= 0) {
-      TRACE(PREFIX_I "Opened device %s\n", device_file);
+      ALLEGRO_DEBUG("Opened device %s\n", device_file);
       /* The device is a mouse if it has a BTN_MOUSE */
       if (has_event(fd, EV_KEY, BTN_MOUSE)) {
-	 TRACE(PREFIX_I "Device %s was a mouse.\n", device_file);
+	 ALLEGRO_DEBUG("Device %s was a mouse.\n", device_file);
       }
       else {
-	 TRACE(PREFIX_I "Device %s was not mouse, closing.\n", device_file);
+	 ALLEGRO_DEBUG("Device %s was not mouse, closing.\n", device_file);
 	 close(fd);
 	 fd = -1;
       }
@@ -564,60 +529,57 @@ static int open_mouse_device (const char *device_file)
  *  Here we open the mouse device, initialise anything that needs it, 
  *  and chain to the framework init routine.
  */
-static int mouse_init (void)
+static bool mouse_init (void)
 {
+/*
    char tmp1[128], tmp2[128];
-   AL_CONST char *udevice;
+   const char *udevice;
+*/
 
    /* Set the current tool */
    current_tool = default_tool;
 
-   /* Find the device filename */
-   udevice = get_config_string (uconvert_ascii ("mouse", tmp1),
-                                uconvert_ascii ("mouse_device", tmp2),
-                                NULL);
+   /* try several /dev/input/event<n>
+    * devices. */
+   const char *device_name[] = { "/dev/input/event0",
+                                 "/dev/input/event1",
+                                 "/dev/input/event2",
+                                 "/dev/input/event3",
+                                 NULL };
+   int i;
 
-   /* Open mouse device.  Devices are cool. */
-   if (udevice) {
-      TRACE(PREFIX_I "Trying %s device\n", udevice);
-      intdrv.device = open_mouse_device (uconvert_toascii (udevice, tmp1));
-      if (intdrv.device < 0) {
-         uszprintf (allegro_error, ALLEGRO_ERROR_SIZE, get_config_text ("Unable to open %s: %s"),
-                    udevice, ustrerror (errno));
-         return -1;
-      }
-   }
-   else {
-      /* If not specified in the config file, try several /dev/input/event<n>
-       * devices. */
-      const char *device_name[] = { "/dev/input/event0",
-                                    "/dev/input/event1",
-                                    "/dev/input/event2",
-                                    "/dev/input/event3",
-                                    NULL };
-      int i;
+   ALLEGRO_DEBUG("Trying /dev/input/event[0-3] devices\n");
 
-      TRACE(PREFIX_I "Trying /dev/input/event[0-3] devices\n");
-
-      for (i=0; device_name[i]; i++) {
-         intdrv.device = open_mouse_device (device_name[i]);
-         if (intdrv.device >= 0) {
+   for (i=0; device_name[i]; i++) {
+      the_mouse.fd = open_mouse_device (device_name[i]);
+      if (the_mouse.fd >= 0)
 	    break;
-         }
-      }
-
-      if (!device_name[i]) {
-	 uszprintf (allegro_error, ALLEGRO_ERROR_SIZE, get_config_text ("Unable to open a mouse device: %s"),
-		    ustrerror (errno));
-	 return -1;
-      }
    }
 
-   intdrv.num_buttons = get_num_buttons(intdrv.device);
-   /* Init the tablet data */
-   init_tablet(intdrv.device);
+   if (!device_name[i]) {
+	 return false;
+   }
 
-   return __al_linux_mouse_init (&intdrv);
+   /* Init the tablet data */
+   init_tablet(the_mouse.fd);
+
+   /* Initialise the mouse object for use as an event source. */
+   _al_event_source_init(&the_mouse.parent.es);
+
+   /* Start watching for data on the fd. */
+   _al_unix_start_watching_fd(the_mouse.fd, process_new_data, &the_mouse);
+
+#ifdef ALLEGRO_RASPBERRYPI
+   ALLEGRO_SYSTEM *s = al_get_system_driver();
+   if (s && s->displays._size > 0) {
+      ALLEGRO_DISPLAY *d = _al_vector_ref(&s->displays, 0);
+      if (d) {
+         _al_evdev_set_mouse_range(0, 0, d->w-1, d->h-1);
+      }
+   }
+#endif
+
+   return true;
 }
 
 
@@ -627,108 +589,273 @@ static int mouse_init (void)
  */
 static void mouse_exit (void)
 {
-   __al_linux_mouse_exit();
-   close (intdrv.device);
+   _al_unix_stop_watching_fd(the_mouse.fd);
+
+   _al_event_source_free(&the_mouse.parent.es);
+
+   close(the_mouse.fd);
+
+   /* This may help catch bugs in the user program, since the pointer
+    * we return to the user is always the same.
+    */
+   memset(&the_mouse, 0, sizeof the_mouse);
+   the_mouse.fd = -1;
 }
 
 
 
-/* mouse_position:
+/* mouse_get_mouse:
+ *  Returns the address of a ALLEGRO_MOUSE structure representing the mouse.
  */
-static void mouse_position(int x, int y)
+static ALLEGRO_MOUSE *mouse_get_mouse(void)
 {
-   DISABLE();
-
-   x_axis.out_abs = x;
-   y_axis.out_abs = y;
-   x_axis.mickeys = 0;
-   y_axis.mickeys = 0;
-
-   _mouse_x = x;
-   _mouse_y = y;
-
-   ENABLE();
+   return (ALLEGRO_MOUSE *)&the_mouse;
 }
 
 
 
-/* mouse_set_range:
+/* mouse_get_mouse_num_buttons:
+ *  Return the number of buttons on the mouse.
  */
-static void mouse_set_range(int x1, int y1, int x2, int y2)
+static unsigned int mouse_get_mouse_num_buttons(void)
 {
-   x_axis.out_min = x1;
-   y_axis.out_min = y1;
-   x_axis.out_max = x2;
-   y_axis.out_max = y2;
-   /* There's no set range for z */
+   ASSERT(the_mouse.fd >= 0);
 
-   DISABLE();
-
-   _mouse_x = CLAMP(x1, _mouse_x, x2);
-   _mouse_y = CLAMP(y1, _mouse_y, y2);
-
-   ENABLE();
+   return get_num_buttons(the_mouse.fd);
 }
 
 
-/* mouse_set_speed:
+
+/* mouse_get_mouse_num_axes:
+ *  Return the number of axes on the mouse.
+ */
+static unsigned int mouse_get_mouse_num_axes(void)
+{
+   ASSERT(the_mouse.fd >= 0);
+
+   /* XXX get the right number later */
+   return 3;
+}
+
+
+
+/* mouse_set_mouse_xy:
+ *
+ */
+static bool mouse_set_mouse_xy(ALLEGRO_DISPLAY *display, int x, int y)
+{
+   (void)display;
+
+   _al_event_source_lock(&the_mouse.parent.es);
+   {
+      int dx, dy;
+
+      x_axis.out_abs = _ALLEGRO_CLAMP(x_axis.out_min, x, x_axis.out_max);
+      y_axis.out_abs = _ALLEGRO_CLAMP(y_axis.out_min, y, y_axis.out_max);
+      x_axis.mickeys = 0;
+      y_axis.mickeys = 0;
+
+      dx = x_axis.out_abs - the_mouse.state.x;
+      dy = y_axis.out_abs - the_mouse.state.y;
+
+      if ((dx != 0) && (dy != 0)) {
+         the_mouse.state.x = x_axis.out_abs;
+         the_mouse.state.y = y_axis.out_abs;
+
+         generate_mouse_event(
+            ALLEGRO_EVENT_MOUSE_AXES,
+            the_mouse.state.x, the_mouse.state.y, the_mouse.state.z,
+            dx, dy, 0,
+            0);
+      }
+   }
+   _al_event_source_unlock(&the_mouse.parent.es);
+
+   return true;
+}
+
+
+
+/* mouse_set_mouse_axis:
+ *
  *   Number of mickeys to cross the screen horizontally: speed * 320.
  */
-static void mouse_set_speed(int speedx, int speedy)
+static bool mouse_set_mouse_axis(int which, int z)
 {
-   float scale = 1;
+   if (which != 2) {
+      return false;
+   }
 
-   if (gfx_driver)
-      scale = gfx_driver->w/320;
+   _al_event_source_lock(&the_mouse.parent.es);
+   {
+      int dz;
 
-   DISABLE();
+      z_axis.out_abs = z;
 
-   x_axis.speed = scale / MAX(1, speedx);
-   y_axis.speed = scale / MAX(1, speedy);
+      dz = z_axis.out_abs - the_mouse.state.z;
 
-   x_axis.mickeys = 0;
-   y_axis.mickeys = 0;
+      if (dz != 0) {
+         the_mouse.state.z = z_axis.out_abs;
 
-   ENABLE();
+         generate_mouse_event(
+            ALLEGRO_EVENT_MOUSE_AXES,
+            the_mouse.state.x, the_mouse.state.y, the_mouse.state.z,
+            0, 0, dz,
+            0);
+      }
+   }
+   _al_event_source_unlock(&the_mouse.parent.es);
+
+   return true;
 }
 
 
 
-/* mouse_get_mickeys:
+/* mouse_set_mouse_range:
+ *
  */
-static void mouse_get_mickeys(int *mickeyx, int *mickeyy)
+bool _al_evdev_set_mouse_range(int x1, int y1, int x2, int y2)
 {
-   int mx = x_axis.mickeys;
-   int my = y_axis.mickeys;
+   _al_event_source_lock(&the_mouse.parent.es);
+   {
+      int dx, dy;
 
-   x_axis.mickeys -= mx;
-   y_axis.mickeys -= my;
+      x_axis.out_min = x1;
+      y_axis.out_min = y1;
+      x_axis.out_max = x2;
+      y_axis.out_max = y2;
 
-   *mickeyx = mx;
-   *mickeyy = my;
+      x_axis.out_abs = _ALLEGRO_CLAMP(x_axis.out_min, x_axis.out_abs, x_axis.out_max);
+      y_axis.out_abs = _ALLEGRO_CLAMP(y_axis.out_min, y_axis.out_abs, y_axis.out_max);
+
+      dx = x_axis.out_abs - the_mouse.state.x;
+      dy = y_axis.out_abs - the_mouse.state.y;
+
+      if ((dx != 0) && (dy != 0)) {
+         the_mouse.state.x = x_axis.out_abs;
+         the_mouse.state.y = y_axis.out_abs;
+
+         generate_mouse_event(
+            ALLEGRO_EVENT_MOUSE_AXES,
+            the_mouse.state.x, the_mouse.state.y, the_mouse.state.z,
+            dx, dy, 0,
+            0);
+      }
+   }
+   _al_event_source_unlock(&the_mouse.parent.es);
+
+   return true;
 }
 
 
 
-
-MOUSE_DRIVER mousedrv_linux_evdev =
+/* mouse_get_state:
+ *  Copy the current mouse state into RET_STATE, with any necessary locking.
+ */
+static void mouse_get_state(ALLEGRO_MOUSE_STATE *ret_state)
 {
-   MOUSEDRV_LINUX_EVDEV,
-   empty_string,
-   empty_string,
+   _al_event_source_lock(&the_mouse.parent.es);
+   {
+      *ret_state = the_mouse.state;
+   }
+   _al_event_source_unlock(&the_mouse.parent.es);
+}
+
+
+
+/* process_new_data: [fdwatch thread]
+ *  Process new data arriving in the keyboard's fd.
+ */
+static void process_new_data(void *data)
+{
+   ASSERT((AL_MOUSE_EVDEV *)data == &the_mouse);
+   
+   _al_event_source_lock(&the_mouse.parent.es);
+   {
+      struct input_event events[32];
+      int bytes, nr, i;
+
+      while ((bytes = read(the_mouse.fd, &events, sizeof events)) > 0) {
+         nr = bytes / sizeof(events[0]);
+         for (i = 0; i < nr; i++) {
+            process_event(&events[i]);
+         }
+      }
+   }
+   _al_event_source_unlock(&the_mouse.parent.es);
+}
+
+
+
+/* [fdwatch thread] */
+static void process_event(struct input_event *event)
+{
+   switch (event->type) {
+      case EV_KEY:
+         process_key(event);
+         break;
+      case EV_REL:
+         process_rel(event);
+         break;
+      case EV_ABS:
+         process_abs(event);
+         break;
+   }
+}
+
+
+
+/* [fdwatch thread] */
+static void generate_mouse_event(unsigned int type,
+                                 int x, int y, int z,
+                                 int dx, int dy, int dz,
+                                 unsigned int button)
+{
+   ALLEGRO_EVENT event;
+
+   if (!_al_event_source_needs_to_generate_event(&the_mouse.parent.es))
+      return;
+
+   event.mouse.type = type;
+   event.mouse.timestamp = al_get_time();
+   event.mouse.display = NULL;
+   event.mouse.x = x;
+   event.mouse.y = y;
+   event.mouse.z = z;
+   event.mouse.dx = dx;
+   event.mouse.dy = dy;
+   event.mouse.dz = dz;
+   event.mouse.button = button;
+   event.mouse.pressure = 0.0; /* TODO */
+   _al_event_source_emit_event(&the_mouse.parent.es, &event);
+}
+
+
+
+/* the driver vtable */
+ALLEGRO_MOUSE_DRIVER _al_mousedrv_linux_evdev =
+{
+   AL_MOUSEDRV_LINUX_EVDEV,
+   "",
+   "",
    "Linux EVDEV mouse (and tablet)",
    mouse_init,
    mouse_exit,
-   NULL, /* poll() */
-   NULL, /* timer_poll() */
-   mouse_position,
-   mouse_set_range,
-   mouse_set_speed,
-   mouse_get_mickeys,
-   analyse_data,
-   NULL, /* enable_hardware_cursor */
-   NULL
+   mouse_get_mouse,
+   mouse_get_mouse_num_buttons,
+   mouse_get_mouse_num_axes,
+   mouse_set_mouse_xy,
+   mouse_set_mouse_axis,
+   mouse_get_state
 };
 
-#endif
 
+
+#endif /* ALLEGRO_HAVE_LINUX_INPUT_H */
+
+/*
+ * Local Variables:
+ * c-basic-offset: 3
+ * indent-tabs-mode: nil
+ * End:
+ */
